#ifndef ROBOT_H
#define ROBOT_H

#include <math.h>
#include <vector>
#include <string>
#include <list>
#include <body.h>
#include <joint.h>
#include <camera.h>



/**
  class Robot
*/
class Robot
{
  private:
    //! name of the file (XML) describing the robot (if it was read from a file)
    /*! NB: the file is given in absolute path   */
    std::string _file;
    std::string _path;
    std::string _name; /*!< name of the robot */

  public:    
    std::vector<Body*> bodies; /*!< array of pointers to the robot's bodies */

    std::vector<Joint*> joints; /*!< array of the robot's joints */
 
    std::vector<Constraint> constraints; /*!< array of the robot's joint constraints */

    std::string models_directory; /*!< name of the directory that contains the 3D models of the robot's bodies */

    Robot();
    ~Robot();

    std::string file();
    std::string path();
    std::string name();
    unsigned int nbBodies();
    unsigned int nbJoints();
    void rename(const std::string name);

    int loadFile(std::string path, std::string filename);
    int draw(bool smooth= true);
    int drawInBlack();
    int drawJoints(float length);
    void printInfo();
    int computeJointRelativePoses();
    int setConfiguration(std::vector<float> &q);
    int getConfiguration(std::vector<float> &q);
    int getVelocity(std::vector<float> &q);
    int getAcceleration(std::vector<float> &q);
    int selectDOF(unsigned int index);
    int setDOFvalue(unsigned int index, float x);
    int setDOFvelocity(unsigned int index, float x);
    int setDOFacceleration(unsigned int index, float x);
    int getDOFvalue(unsigned int index, float &x);
    int incrementDOFvalue(unsigned int index, float ratio);
    int decrementDOFvalue(unsigned int index, float ratio);
    int getDOFunit(unsigned int index, DOFunit &unit);
    int getDOFs(std::vector<DOF> &q);
    int getDOFbounds(unsigned int index, float &min, float &max);
    unsigned int nbDOFs();
    void updatePassiveJoints();
    void updateConfiguration();
    int printConfiguration(DOFunit angle_unit);
    int addJoint(std::string _name, JointType type, std::vector<float> qmin, std::vector<float> qmax, HTMatrix Tconst, std::string previous);
    int addBody(std::string _name, BodyType type, std::string file, float scale, std::string joint_name, HTMatrix _Tjoint);
    int addBody(Body *body, std::string joint_name, HTMatrix _Tjoint);
    int addConstraint(std::string master, std::string slave, ConstraintType type, unsigned int master_dof, unsigned int slave_dof, std::vector<float> params);
    bool selfCollisionTest();
    float selfCollisionDistance(Vector3D &closest_point1, Vector3D &closest_point2);
    bool motionSelfCollisionTest(std::vector<float> q, std::vector<float> speed, float time);
    bool bodyCollisionTest(Body *obstacle);
    int reset();
    Body *getBodyByName(std::string _name);
    Joint *getJointByName(std::string _name);
    int connectRobot(std::string connection, HTMatrix &T, Robot *subrobot);
    int exportBodyModelsInJointFrames();
    int enableDisplayOfCollisionBodiesOnly();
    int resetDisplayStates();
    int enableDisplay(bool on);
    int enableCollision(bool on);
};



/*! \brief This class contains the robot and the obstacle array. */
class World
{
  private:
    //! name of the file (XML) describing the world (if it was read from a file)
    /*! NB: the file is given in absolute path   */
    std::string _file;
    std::string _path;

  public:
    Robot robot;
    std::vector<Body*> obstacles;
    std::string models_directory; /*!< name of the directory that contains the 3D models of the world obstacles */
    std::string robot_directory; /*!< name of the directory that contains the XML file describing the robot */
    float xmin;  /*!< minimal limit of the world bounding box along X-axis */
    float xmax;  /*!< maximal limit of the world bounding box along X-axis */
    float ymin;  /*!< minimal limit of the world bounding box along Y-axis */
    float ymax;  /*!< maximal limit of the world bounding box along Y-axis */
    float zmin;  /*!< minimal limit of the world bounding box along Z-axis */
    float zmax;  /*!< maximal limit of the world bounding box along Z-axis */

    Vector3D lightPosition; /*!< for OpenGL and POVRay export */

    World();
    ~World();

    std::string file()
    {  return _file;  }
    std::string path()
    {  return _path;  }

    int draw(bool smooth= true);
    int drawInBlack();
    int addObstacle(std::string name, BodyType type, std::string file, float scale, HTMatrix _pose);
    int addObstacle(Body *body, HTMatrix _pose);
    int removeObstacles(const std::string &prefix);
    int loadFile(std::string path, std::string filename);
    int reset();
    bool collisionTest();
    bool earlyCollisionTest(float time);
    float closestCollision(Vector3D &closestPoint1, Vector3D &closestPoint2);
    int displayObstacleTextures(bool on);
#if defined(ENABLE_DISPLAY)
    int exportSceneToPOVRay(Camera &camera, std::string foldername, std::string filename, bool highQuality= false);
#endif
    int enableDisplayOfCollisionBodiesOnly();
    int resetDisplayStates();
};

#endif
